module myproblem

use globals, only: problem,linesearch,Aparam,Bparam,Cparam,Dparam,detx,xinv

implicit none

  ! SUBROUTINES
  public :: inip, evalf, evalg, expRiem

contains

	subroutine inip(n,m,x,scaleF)

	implicit none

	! SCALAR ARGUMENTS
	integer, intent(out) :: n,m
	logical, intent(out) :: scaleF
	
	! ARRAY ARGUMENTS
	real(kind=8), allocatable, intent(out) :: x(:,:)
	
	! LOCAL SCALARS
	integer :: i,j,allocerr,info,iseed(4)
	real(kind=8) :: t1,t2
	
	! LOCAL ARRAYS
	real(kind=8), allocatable :: work(:)
	
	! FUNCTIONS
	real(kind=8) :: dlaran
	

	! ----------------------------------------------------------------------
	
	! 
	
	if ( problem == 1 ) then 
	
			! Number of variables
			
			n = 3
			
			! Number of objectives
			
			m = 2
			
			allocate(x(n,n),Aparam(m),Bparam(m),work(3*n),stat=allocerr)
			if ( allocerr .ne. 0 ) then
				 write(*,*) 'Allocation error in main program'
				 stop
			end if
			
			! Set Aparam and Bparam		
			
			iseed(1) = 2
			iseed(2) = 3
			iseed(3) = 4
			iseed(4) = 9
			
			t1 = 0.0d0
			t2 = 1.0d0
			
			do i = 1,m
					Aparam(i) = t1 + ( t2 - t1 ) * dlaran(iseed)
					Bparam(i) = t1 + ( t2 - t1 ) * dlaran(iseed)
			end do
			
			! Initial point
			
			t1 = 0.0d0
			t2 = 2.0d1
			
			x(:,:) = 0.0d0
			do i = 1,n
					x(i,i) = t1 + ( t2 - t1 ) * dlaran(iseed)
			end do 		

			call dlaror('C','N',n,n,x,n,iseed,work,info)
			

			! Define the line search to be used
			! linesearch = 1: Armijo line search
			! linesearch = 2: constant step size
			! linesearch = 3: hybrid line search
			
			linesearch = 1
					
			! Scale the problem?
			
			scaleF   = .false.
			
			
			return
	end if		
	
	! ----------------------------------------------------------------------
		
	if ( problem == 2 ) then 
	
			! Number of variables
			
			n = 3
			
			! Number of objectives
			
			m = 2
			
			allocate(x(n,n),Aparam(m),Bparam(m),Cparam(m),Dparam(m),work(3*n),stat=allocerr)
			if ( allocerr .ne. 0 ) then
				 write(*,*) 'Allocation error in main program'
				 stop
			end if
			
			! Set Aparam and Bparam		
			
			iseed(1) = 2
			iseed(2) = 3
			iseed(3) = 4
			iseed(4) = 9
			
			t1 = 0.0d0
			t2 = 1.0d0
			
			do i = 1,m
					Aparam(i) = t1 + ( t2 - t1 ) * dlaran(iseed)
					Bparam(i) = t1 + ( t2 - t1 ) * dlaran(iseed)
					Cparam(i) = t1 + ( t2 - t1 ) * dlaran(iseed)
					Dparam(i) = Aparam(i) * Bparam(i) * dlaran(iseed)
			end do
			
			! Initial point
			
			t1 = 0.0d0
			t2 = 1.0d2
			
			x(:,:) = 0.0d0
			do i = 1,n
					x(i,i) = t1 + ( t2 - t1 ) * dlaran(iseed)
			end do 		

			call dlaror('C','N',n,n,x,n,iseed,work,info)
			
			! Define the line search to be used
			! linesearch = 1: Armijo line search
			! linesearch = 2: constant step size
			! linesearch = 3: hybrid line search
			
			linesearch = 1
					
			! Scale the problem?
			
			scaleF   = .false.
			
			
			return
	end if		
	
	end subroutine inip

	!***********************************************************************
	!***********************************************************************

	subroutine evalf(n,x,f,ind)
	
	implicit none

	! SCALAR ARGUMENTS
	integer, intent(in) :: n,ind
	real(kind=8), intent(out) :: f
	
	! ARRAY ARGUMENTS
	real(kind=8), intent(in) :: x(n,n)
	
	! LOCAL SCALARS
	integer :: i,inform
	real(kind=8) :: t,U(n,n)
	
! ----------------------------------------------------------------------

	! 
			
	if ( problem == 1 ) then 
			
		U = x

		call DPOTRF('L',n,U,n,inform) 
		
		if ( inform /= 0 ) then
			write(*,*) 'An error occurred in the Cholesky factorization process'
			return
		end if
	
		! Compute the determinant of x
	
		t = U(1,1)
		do i = 2,n
			t = t * U(i,i)
		end do
		t = t ** 2

		t = log( t )

		f = Aparam(ind) * t ** 2 - Bparam(ind) * t
	
		return
	end if	
	
! ----------------------------------------------------------------------	
	
	! Example 23
			
	if ( problem == 2 ) then 
			
		U = x

		call DPOTRF('L',n,U,n,inform) 
		
		if ( inform /= 0 ) then
			write(*,*) 'An error occurred in the Cholesky factorization process'
			return
		end if
	
		! Compute the determinant of x
	
		t = U(1,1)
		do i = 2,n
			t = t * U(i,i)
		end do
		t = t ** 2

		f = Aparam(ind) * log( t ** Bparam(ind) + Cparam(ind) ) - Dparam(ind) * log( t )
	
		return
	end if		
						
	end subroutine evalf

	!***********************************************************************
	!***********************************************************************

	subroutine evalg(n,x,g,ind)
	
	implicit none
	
	! SCALAR ARGUMENTS
	integer, intent(in) :: n,ind
	
	! ARRAY ARGUMENTS
	real(kind=8), intent(in)  :: x(n,n)
	real(kind=8), intent(out) :: g(n,n)
	
	! LOCAL SCALARS
	
	integer :: i,inform
	real(kind=8) :: t,U(n,n)
	
! ----------------------------------------------------------------------

	! 

	if ( problem == 1 ) then 
			
			g = ( 2.0d0 * Aparam(ind) * log( detx ) - Bparam(ind) ) * xinv
						
			return
			
	end if	
	
! ----------------------------------------------------------------------			
	
	! Example 23

	if ( problem == 2 ) then 
			
			g = ( Aparam(ind) * Bparam(ind) * detx ** Bparam(ind) /         &
			      ( detx ** Bparam(ind) + Cparam(ind) ) - Dparam(ind) ) * xinv
						
			return
			
	end if

		
				

	end subroutine evalg
	

	
	!***********************************************************************
	!***********************************************************************
	
	subroutine expRiem(n,x,u,expR,inform)
	
	implicit none
	
	! SCALAR ARGUMENTS
	integer, intent(in)  :: n
	integer, intent(out) :: inform
	
	! ARRAY ARGUMENTS
	real(kind=8), intent(in)  :: x(n,n),u(n,n)
	real(kind=8), intent(out) :: expR(n,n)	

	! LOCAL SCALARS
	integer :: i,j,iexph,ns,ideg,lwsp,info,allocerr
	
	! LOCAL ARRAYS
	integer :: ipiv(n)
	real(kind=8) :: expD(n),L(n,n),aux(n,n)
	real(kind=8), allocatable :: wsp(:)
	
	ideg = 6
	lwsp = 4*n*n+ideg+1
	
	allocate(wsp(lwsp),stat=allocerr)
	if ( allocerr .ne. 0 ) then
		write(*,*) 'Allocation error in main program'
		stop
	end if
	
! ----------------------------------------------------------------------

	! 
	
	if ( problem == 1 .or. problem == 2 ) then 
			
			expR = matmul( xinv, u )
			
			call DGPADM(ideg,n,1.0d0,expR,n,wsp,lwsp,ipiv,iexph,ns,info )
						
			if ( INFO /= 0 ) then
				write(*,*) 'An error occurred in the exp process'
				inform = -1
				return
			end if
		
			do i = 1,n
				do j = 1,n
					expR(i,j) = wsp(iexph+(j-1)*n+i-1)
				end do
			end do
			
			expR = matmul(x,expR)
					
			expR = ( expR + transpose(expR) ) / 2.0d0
						
			inform = 0
			
			return
	end if		
	
	end subroutine expRiem
	
!***********************************************************************
!***********************************************************************
	
	subroutine Lipschitz(n,L,inform)
	
	implicit none
	
	! SCALAR ARGUMENTS
	integer,intent(in) :: n
	integer,intent(out) :: inform
	real(kind=8), intent(out) :: L
	
	inform = -1
	
	if ( problem == 1 ) then 
			
			L = 2.0d0 * sqrt(real(n)) * maxval(Aparam)
			
			inform = 0
			
			return
	end if
	
! ----------------------------------------------------------------------	
	
	if ( problem == 2 ) then 
	
			L = n * maxval( Aparam * Bparam**2 )
			
			inform = 0
			
			return
	end if
	
	end subroutine Lipschitz

end module myproblem
