module myproblem

use globals, only: problem,seed,Aparam,Bparam,Cparam,Uparam,Wparam,Riemmanian

implicit none

  ! SUBROUTINES
  public :: inip, evalf, evalg, evalGmetric, expRiem

contains

	subroutine inip(n,m,x,scaleF,checkder)

	implicit none

	! SCALAR ARGUMENTS
	integer, intent(out) :: n,m
	logical, intent(out) :: scaleF,checkder
	
	! ARRAY ARGUMENTS
	real(kind=8), allocatable, intent(out) :: x(:)
	
	! LOCAL SCALARS
	integer :: i,j,allocerr
	real(kind=8) :: a,b
	
	! FUNCTIONS
	real(kind=8) :: drand
	

! ----------------------------------------------------------------------

	! 	Rosembrock
	
	if ( problem == 1 ) then 
	
			! Number of variables
	
			n = 2
			
			! Number of objectives
			
			m = 2
			
			allocate(x(n),stat=allocerr)
			if ( allocerr .ne. 0 ) then
				 write(*,*) 'Allocation error in main program'
				 stop
			end if
			
			! Initial point

			a = -5.0d0
			b =  5.0d0
			
		    seed = 123456.0d0

			
			do i = 1,n
					x(i) = a + ( b - a ) * drand(seed)
			end do 
						
			! Scale the problem?
			
			scaleF   = .false.
			
			! Check derivatives?
			
			checkder = .false.
	end if		
	
! ----------------------------------------------------------------------

	! 	Example 30
	
	if ( problem == 2 ) then 
	
			! Number of variables
	
			n = 100
			
			! Number of objectives
			
			m = 2
			
			allocate(x(n),Aparam(m),Bparam(m),Cparam(m),Uparam(n,m),Wparam(n,m),stat=allocerr)
			if ( allocerr .ne. 0 ) then
				 write(*,*) 'Allocation error in main program'
				 stop
			end if
			
			! Initial point

			a =  0.0d0
			b =  1.0d1
			
			seed = 123456.0d0
			
			do i = 1,n
					x(i) = a + ( b - a ) * drand(seed)
			end do 
						
			! Scale the problem?
			
			scaleF   = .false.
			
			! Check derivatives?
			
			checkder = .false.
			
			! Define the parameters of the problem
			
			a =  0.0d0
			b =  1.0d0
			
			do j = 1,m
					Aparam(j) = a + ( b - a ) * drand(seed)
					Bparam(j) = a + ( b - a ) * drand(seed)
					Cparam(j) = a + ( b - a ) * drand(seed)
					
					do i = 1,n
							Uparam(i,j) = a + ( b - a ) * drand(seed)
							Wparam(i,j) = a + ( b - a ) * drand(seed)					
					end do
					
			end do 
			
	end if					
	
	
	end subroutine inip

	!***********************************************************************
	!***********************************************************************

	subroutine evalf(n,x,f,ind)
	
	implicit none

	! SCALAR ARGUMENTS
	integer, intent(in) :: n,ind
	real(kind=8), intent(out) :: f
	
	! ARRAY ARGUMENTS
	real(kind=8), intent(in) :: x(n)
	
	! LOCAL SCALARS
	integer :: i
	real(kind=8) :: a,b
		
! ----------------------------------------------------------------------

	! 	Rosenbrock
	
	if ( problem == 1 ) then 
			
			if ( ind == 1 ) then
					a = 1.0d2
					b = 1.0d0
					
					f = 0.0d0
					do i = 1,n/2
							f = f + a * ( x(2*i-1) ** 2 - x(2*i) ) ** 2 + ( x(2*i-1) - b ) ** 2
					end do		
					return
			end if
			
			if ( ind == 2 ) then
					a = 1.0d2
					b = 2.0d0
					
					f = 0.0d0
					do i = 1,n/2
							f = f + a * ( x(2*i-1) ** 2 - x(2*i) ) ** 2 + ( x(2*i-1) - b ) ** 2
					end do		
					return
			end if
			
	end if			
	
! ----------------------------------------------------------------------

	! 	Example 30
	
	if ( problem == 2 ) then 
	
			f = Aparam(ind) * log( product( x ** Uparam(:,ind) ) + Bparam(ind) ) - &
			    sum( Wparam(:,ind) * log( x ) ) + Cparam(ind) * sum( log( x ) ** 2  )
			    
			return    
					
	end if				
	
					
	end subroutine evalf

	!***********************************************************************
	!***********************************************************************

	subroutine evalg(n,x,g,ind)
	
	implicit none
	
	! SCALAR ARGUMENTS
	integer, intent(in) :: n,ind
	
	! ARRAY ARGUMENTS
	real(kind=8), intent(in)  :: x(n)
	real(kind=8), intent(out) :: g(n)
	
	! LOCAL SCALARS
	
	integer :: i
	real(kind=8) :: a,b,aux
		
	! ----------------------------------------------------------------------

	! 	Rosenbrock
	
	if ( problem == 1 ) then 
				
			if ( ind == 1 ) then
					a = 1.0d2
					b = 1.0d0
									
					do i = 1,n/2
							g(2*i-1) = 4.0d0 * a * ( x(2*i-1) ** 2 - x(2*i) ) * x(2*i-1) + &
							           2.0d0 * ( x(2*i-1) - b )
							g(2*i)   = - 2.0d0 * a * ( x(2*i-1) ** 2 - x(2*i) )
					end do
					return
			end if
			
			if ( ind == 2 ) then
					a = 1.0d2
					b = 2.0d0
					
					do i = 1,n/2
							g(2*i-1) = 4.0d0 * a * ( x(2*i-1) ** 2 - x(2*i) ) * x(2*i-1) + &
												 2.0d0 * ( x(2*i-1) - b )
							g(2*i)   = - 2.0d0 * a * ( x(2*i-1) ** 2 - x(2*i) )
					end do
					return
			end if
			
	end if
	
	
! ----------------------------------------------------------------------

	! 	Example 30
	
	if ( problem == 2 ) then 
	
			aux = product( x ** Uparam(:,ind) )
			
			do i = 1,n
					g(i) = Aparam(ind) * Uparam(i,ind) * aux / ( x(i) * ( aux + Bparam(ind) ) ) &
					       - Wparam(i,ind) / x(i) + 2.0d0 * Cparam(ind) * log( x(i) ) / x(i)
			end do
				
			return
	end if	
					

	end subroutine evalg
	

	!***********************************************************************
	!***********************************************************************
	
	subroutine evalGmetric(n,x,G)
	
	implicit none
	
	! SCALAR ARGUMENTS
	integer, intent(in) :: n
	real(kind=8), intent(out) :: G(n,n)	
	
	! ARRAY ARGUMENTS
	real(kind=8), intent(in)  :: x(n)
	
	! LOCAL ARRAYS
	integer :: i
	
	! < u ,v > = u^T G(x) v
	

	! 	Rosenbrock
	
	if ( problem == 1 ) then 
	
			if ( Riemmanian ) then
			
					G(:,:) = 0.0d0
					
					do i = 1,n/2
							G(2*i-1,2*i-1) = 1.0d0 + 4.0d0 * x(2*i-1) ** 2
							G(2*i,2*i-1)   = - 2.0d0 * x(2*i-1)
							G(2*i-1,2*i)   = G(2*i,2*i-1)
							G(2*i,2*i)     = 1.0d0
					end do
			else
									
					G(:,:) = 0.0d0 
					forall(i = 1:n) G(i,i) = 1.0d0 
					
			end if
	
			return
	end if		
	
! ----------------------------------------------------------------------

	! 	Example 30
	
	if ( problem == 2 ) then 
	
			if ( Riemmanian ) then
					G(:,:) = 0.0d0
					
					do i = 1,n
							G(i,i) = 1.0d0 / x(i) ** 2
					end do
			else
					
		 			G(:,:) = 0.0d0 
					forall(i = 1:n) G(i,i) = 1.0d0 
			end if
						
			return
	end if
	
	
	
		
	end subroutine evalGmetric
	
	!***********************************************************************
	!***********************************************************************
	
	subroutine expRiem(n,x,u,expR)
	
	implicit none
	
	! SCALAR ARGUMENTS
	integer, intent(in) :: n
	
	! ARRAY ARGUMENTS
	real(kind=8), intent(in)  :: x(n),u(n)
	real(kind=8), intent(out) :: expR(n)	

	
	! LOCAL ARRAYS
	integer :: i
	
	
! ----------------------------------------------------------------------

	! 	Rosenbrock
	
	if ( problem == 1 ) then 
	
			if ( Riemmanian ) then
	
					do i = 1,n/2
							expR(2*i-1) = x(2*i-1) + u(2*i-1)
							expR(2*i)   = x(2*i) + u(2*i-1) ** 2 + u(2*i)
					end do
			else
			
					expR = x + u
			end if

			return
			
	end if		
	
! ----------------------------------------------------------------------

	! 	Example 30
	
	if ( problem == 2 ) then 
	
			if ( Riemmanian ) then
					
					do i = 1,n
							expR(i) = x(i) * exp( u(i) / x(i) )	
					end do
			else
					
					expR = x + u
			end if
			return
	end if
	
			
	
	end subroutine expRiem
	

end module myproblem
